/* ----------------------------------------------------------------------------
 * Lagrange interpolation suitable for NOAA level 1B file formats.
 * Submitted by Andrew Brooks <arb@sat.dundee.ac.uk>
 * Version 1.00, 2008-05-26.
 */

/* ----------------------------------------------------------------------------
 * Perform a Lagrangian interpolation through the given x,y coordinates
 * and return the interpolated y value for the given x value.
 * The array size and thus the polynomial order is defined by numpt.
 * Input: x[] and y[] are of size numpt,
 *  x0 is the x value for which we calculate the corresponding y
 * Returns: y value calculated for given x0.
 */
float
lagrange(float x[], float y[], float x0, int numpt)
{
	int i, j;
	float L;
	float y0 = 0;

	for (i=0; i<numpt; i++)
	{
		L = 1.0;
		for (j=0; j<numpt; j++)
		{
			if (i == j)
				continue;
			L = L * (x0 - x[j]) / (x[i] - x[j]);
		}
		y0 = y0 + L * y[i];
	}
	return(y0);
}


/* ----------------------------------------------------------------------------
 * Interpolate an array of size NUM_PIX where the only values set on input are
 * at KNOWN_FIRST, and intervals of KNOWN_STEP thereafter.
 * On return all the rest from 0..NUM_PIX-1 will be filled in.
 * Uses the lagrange() function to do the interpolation; 5-point for the
 * beginning and end of the array and 4-point for the rest.
 * To use this function for NOAA level 1B data extract the 51 latitude values
 * into their appropriate places in the vals array then call n1b_interp to
 * calculate the rest of the values.  Do similarly for longitudes, solar zenith
 * angles, and any others which are present in the file.
 * Reference:
 *  http://www2.ncdc.noaa.gov/docs/klm/html/c2/sec2-4.htm
 */
#define NUM_KNOWN    51  /* Number of known points */
#define KNOWN_FIRST  24  /* Index in full array of first known point */
#define KNOWN_STEP   40  /* Interval to next and subsequent known points */
#define NUM_PIX    2048  /* Number of points in whole array */
#define INTERP_ORDER  5  /* Ensure this is an odd number, 5 is suitable */
/* Convert number of known point to its index in full array */
#define IDX(N) ((N)*KNOWN_STEP+KNOWN_FIRST)
typedef float n1b_interp_t;

void
n1b_interp(n1b_interp_t vals[])
{
	int i, j;
	n1b_interp_t x[INTERP_ORDER];
	n1b_interp_t y[INTERP_ORDER];

	/* First extrapolate first 24 points */
	for (i=0; i<INTERP_ORDER; i++)
	{
		x[i] = IDX(i);
		y[i] = vals[IDX(i)];
	}
	for (i=0; i<KNOWN_FIRST; i++)
	{
		vals[i] = lagrange(x, y, i, INTERP_ORDER);
	}

	/* Next extrapolate last 23 points */
	for (i=0; i<INTERP_ORDER; i++)
	{
		x[i] = IDX(NUM_KNOWN-INTERP_ORDER+i);
		y[i] = vals[IDX(NUM_KNOWN-INTERP_ORDER+i)];
	}
	for (i=IDX(NUM_KNOWN-1); i<NUM_PIX; i++)
	{
		vals[i] = lagrange(x, y, i, INTERP_ORDER);
	}

	/* Interpolate all intermediate points using two before and two after */
	for (i=KNOWN_FIRST; i<IDX(NUM_KNOWN-1); i++)
	{
		n1b_interp_t x[4];
		n1b_interp_t y[4];
		int startpt;

		/* If we already know this value then skip the calculation */
		if (vals[i] != 0.0)
			continue;

		/* Find a suitable set of two known points before and two after */
		startpt = (i/KNOWN_STEP)-2;
		if (startpt<0) startpt=0;
		if (startpt+3 >= NUM_KNOWN) startpt = NUM_KNOWN-4;
		for (j=0; j<4; j++)
		{
			x[j] = IDX(startpt+j);
			y[j] = vals[IDX(startpt+j)];
		}
		vals[i] = lagrange(x, y, i, 4);
	}
}


/* ----------------------------------------------------------------------------
 */
